# Copyright 2023 Inductor, Inc.
"""Functionality for test suite execution tasks."""

from concurrent import futures
import contextlib
import datetime
from typing import Dict, Generator, Iterator, List, Tuple, Union

import inquirer
import rich
from rich import progress
import typer

from inductor import config
from inductor.backend_client import backend_client, wire_model
from inductor.data_model import data_model
from inductor.execution import execution, quality_measure_execution


@contextlib.contextmanager
def _disable_logger_decorator():
    """Disable the inductor.logger decorator within this context manager.

    Disable the inductor.logger decorator by setting
    `execution.logger_decorator_enabled` to False. On exit, restore the
    original value of `execution.logger_decorator_enabled`.
    """
    orig_logger_decorator_enabled = execution.logger_decorator_enabled
    try:
        execution.logger_decorator_enabled = False
        yield
    finally:
        execution.logger_decorator_enabled = orig_logger_decorator_enabled


def _execute_test_case(
    test_suite_run_id: int,
    llm_program_fully_qualified_name: str,
    test_case_with_metadata: data_model.TestCaseWithMetadata,
    test_case_replica_index: int,
    quality_measures_with_metadata: List[
        data_model.QualityMeasureWithMetadata],
    hparams: Dict[str, Union[str, int, float]],
    auth_access_token: str,
) -> Tuple[
    wire_model.LogTestCaseExecutionRequest,
    List[Dict[str, Union[
        int, str, wire_model.QualityMeasureExecutionDetails]]]]:
    """Run a test case and evaluate its quality measures.

    Sends the output of the test case and the outputs of the quality measures
    to the backend server.

    Args:
        test_suite_run_id: ID of test suite run.
        llm_program_fully_qualified_name: Fully qualified name of LLM program.
        test_case_with_metadata: Test case with ID.
        test_case_replica_index: Index of test case replica.
        quality_measures_with_metadata: List of quality measures with ID.
        hparams: Mapping from hyperparameter names to values.
        auth_access_token: Auth0 access token.
    
    Returns:
        A tuple of the LogTestCaseExecutionRequest and a list of invalid
            quality measures if any. Each invalid quality measure is a
            Dictionary with the following keys:
                "id": ID of quality measure.
                "name": Name of quality measure.
                "execution_details": wire_model.QualityMeasureExecutionDetails
                    object.
    """
    started_at = datetime.datetime.now(datetime.timezone.utc)

    llm_program_stdout = None
    llm_program_stderr = None
    llm_program_error = None
    llm_program_output = None
    # Note: we use backslashes to split the `with` clause immediately below,
    # rather than enclosing in parentheses, because enclosing in parentheses
    # apparently causes a SyntaxError when running in Python 3.8.
    # pylint: disable-next=protected-access
    with execution.capture_stdout_stderr(suppress=True) as (stdout, stderr), \
        execution.capture_logged_values() as logged_values, \
        _disable_logger_decorator(), \
        execution.set_hparams(hparams):
        try:
            # Run the LLM program.
            llm_program_output = data_model.LazyCallable(
                llm_program_fully_qualified_name
            )(**test_case_with_metadata.inputs)
            if isinstance(llm_program_output, Iterator):
                llm_program_output = list(llm_program_output)
                if all(isinstance(value, str) for value in llm_program_output):
                    llm_program_output = "".join(llm_program_output)

        except Exception as error:  # pylint: disable=broad-except
            llm_program_error = str(error)

        llm_program_stdout = stdout.getvalue()
        llm_program_stderr = stderr.getvalue()

    ended_at = datetime.datetime.now(datetime.timezone.utc)

    # If the LLM program completed without error, run the executable quality
    # measures.
    if llm_program_error is None:
        direct_evaluations, invalid_quality_measures = (
            quality_measure_execution.execute_quality_measures(
                test_case=test_case_with_metadata,
                llm_program_output=llm_program_output,
                quality_measures_with_metadata=quality_measures_with_metadata
            ))
    else:
        direct_evaluations = []
        invalid_quality_measures = []

    request_object = wire_model.LogTestCaseExecutionRequest(
        test_suite_run_id=test_suite_run_id,
        test_case_id=test_case_with_metadata.id,
        test_case_replica_index=test_case_replica_index,
        execution_details=wire_model.ExecutionDetails(
            mode="CLI",
            inputs=test_case_with_metadata.inputs,
            hparams=hparams or None,
            output=llm_program_output,
            error=llm_program_error,
            stdout=llm_program_stdout,
            stderr=llm_program_stderr,
            execution_time_secs=(ended_at - started_at).total_seconds(),
            started_at=started_at,
            ended_at=ended_at,
            logged_values=logged_values or None,
            direct_evaluations=direct_evaluations or None,
        )
    )

    backend_client.log_test_case_execution(request_object, auth_access_token)
    return request_object, invalid_quality_measures


@contextlib.contextmanager
def _manage_run_requests(
    test_suite: data_model.TestSuite,
    auth_access_token: str
) -> Generator[wire_model.CreateTestSuiteRunResponse, None, None]:
    """Send requests to the server to manage the creation/completion of a run.
    
    Send a request to the server to create a test suite run. Then, yield the
    response, which contains test suite run metadata. On exit, send a request
    to the server to mark the test suite run as complete.

    Args:
        test_suite: Test suite.
        auth_access_token: Auth0 access token.

    Yields:
        CreateTestSuiteRunResponse object.
    """
    test_suite_run = test_suite._get_run_request()  # pylint: disable=protected-access
    test_suite_run_metadata = backend_client.create_test_suite_run(
        test_suite_run, auth_access_token)
    try:
        yield test_suite_run_metadata
    finally:
        backend_client.complete_test_suite_run(
            wire_model.CompleteTestSuiteRunRequest(
                test_suite_run_id=
                    test_suite_run_metadata.test_suite_run_id,
                ended_at=
                    datetime.datetime.now(datetime.timezone.utc)),
            auth_access_token)


def execute_test_suite(
    test_suite: data_model.TestSuite,
    auth_access_token: str,
    *,
    prompt_open_results: bool = False):
    """Execute a test suite.
    
    Execute a test suite while displaying relevant information to the user,
    including a progress bar.
    
    Args:
        test_suite: Test suite to execute.
        auth_access_token: Auth0 access token.
        prompt_open_results: Whether to prompt the user to open the test
            suite run results in a browser.
    """
    # Note: we use backslashes to split the `with` clause immediately below,
    # rather than enclosing in parentheses, because enclosing in parentheses
    # apparently causes a SyntaxError when running in Python 3.8.
    with futures.ProcessPoolExecutor(
            max_workers=test_suite.config.parallelize) as executor, \
        _manage_run_requests(
            test_suite, auth_access_token) as test_suite_run_metadata:

        # Note that `test_suite_run_metadata.test_case_ids` is the IDs
        # of `test_suite.test_cases` followed by the IDs of
        # `test_suite_run_metadata.imported_test_cases`. This ordering
        # needs to be preserved when updating the components with IDs.
        test_cases_with_metadata = data_model.components_with_ids(
            test_suite.test_cases +
            (test_suite_run_metadata.imported_test_cases or []),
            test_suite_run_metadata.test_case_ids,
            data_model.TestCaseWithMetadata)
        quality_measures_with_metadata = data_model.components_with_ids(
            test_suite.quality_measures,
            test_suite_run_metadata.quality_measure_ids,
            data_model.QualityMeasureWithMetadata)

        hparam_combinations = execution.get_hparams_combinations(
            test_suite.hparam_specs)

        def pluralize(count: int) -> str:
            return "s" if count != 1 else ""
        rich.print(f"Running {len(test_cases_with_metadata)} test case"
              f"{pluralize(len(test_cases_with_metadata))} "
              f"({test_suite.config.replicas} run"
              f"{pluralize(test_suite.config.replicas)} each) ...")

        test_case_futures = []
        for test_case_replica_index in range(test_suite.config.replicas):
            for hparams in hparam_combinations:
                for test_case_with_metadata in test_cases_with_metadata:
                    test_case_futures.append(executor.submit(
                        _execute_test_case,
                        test_suite_run_id=
                            test_suite_run_metadata.test_suite_run_id,
                        llm_program_fully_qualified_name=
                            test_suite.config.llm_program,
                        test_case_with_metadata=test_case_with_metadata,
                        test_case_replica_index=test_case_replica_index,
                        quality_measures_with_metadata=
                            quality_measures_with_metadata,
                        hparams=hparams,
                        auth_access_token=auth_access_token,
                    ))

        typer.echo(f"Go to {test_suite_run_metadata.url} to view results.")
        if prompt_open_results:
            open_url = inquirer.confirm(
                message="Open in browser?", default=True)
            if open_url:
                typer.launch(test_suite_run_metadata.url)

        # Display progress bar and optionally print test outputs.
        with progress.Progress() as progress_bar:
            progress_task = progress_bar.add_task(
                "Test Cases",
                total=(test_suite.config.replicas *
                        len(test_cases_with_metadata) *
                        len(hparam_combinations)))

            for future in futures.as_completed(test_case_futures):
                progress_bar.advance(progress_task)
                test_output, invalid_quality_measures = future.result()
                if test_output.execution_details.error is not None:
                    progress_bar.console.print(
                        "\n[red][bold][ERROR] Test case execution "
                        "raised an exception.[/bold] The following "
                        "execution will be recorded as FAILED and "
                        "quality measures will not be evaluated:[/red]")
                    progress_bar.console.print(test_output)
                elif config.verbose:
                    progress_bar.console.print(test_output)
                if invalid_quality_measures:
                    progress_bar.console.print(
                        "\n[red][bold][ERROR] One or more quality "
                        "measures raised an exception or returned an "
                        "invalid value.[/bold] The following quality "
                        "measures will not be recorded as part of this "
                        "test case execution:[/red]")
                    progress_bar.console.print(invalid_quality_measures)

        rich.print("Run complete.")



# Copyright 2023 Inductor, Inc.
"""Data abstractions and utilities for the Inductor client."""

import asyncio
import copy
import importlib
import inspect
import json
import os
import sys
import types
from typing import Any, Callable, Dict, List, Literal, Optional, Type, TypeVar, Union

import pydantic
import yaml

from inductor import auth_session
from inductor.backend_client import backend_client, wire_model


class TestSuiteValueError(ValueError):
    """Error raised when a test suite is invalid.

    Attributes:
        message: Error message.
        path: Optional path to the test suite file that caused the error.
    """
    def __init__(self, message: str, path: Optional[str] = None):
        """Error raised when a test suite is invalid.

        Args:
            message: Error message.
            path: Optional path to the test suite file that caused the error.
        """
        self.message = message
        self.path = path
        # TODO: Add the line and line number from the test suite file that
        # caused the error, if applicable.
        super().__init__(self.message)


def _extract_fully_qualified_name(value: Any) -> Any:
    """Return the fully qualified name from a Callable, if applicable.
    
    Args:
        value: Object from which to extract the fully qualified name.
    
    Returns:
        The fully qualified name of the given object, if it is a Callable.
        Otherwise, the given object.
    """
    if callable(value):
        return f"{get_module_qualname(value)}:{value.__qualname__}"
    return value


# Type variable for the TestCase class.
_T_TestCase = TypeVar("_T_TestCase", bound="TestCase")  # pylint: disable=invalid-name


class TestCase(pydantic.BaseModel):
    """A test case.

    Attributes:
        inputs: Mapping from input parameter name to input value.
        output: Optionally, an example of a desired high-quality output, or
            the output that is to be considered correct, for this test case.
        description: Optionally, a description of this test case.
    """
    model_config = pydantic.ConfigDict(extra="forbid")

    inputs: Dict[str, Any]
    output: Optional[Any] = None
    description: Optional[str] = None

    # Define a custom constructor to enable taking `inputs` as a positional,
    # rather than keyword-only, argument.
    def __init__(
        self,
        inputs: Dict[str, Any],
        *,
        output: Optional[Any] = None,
        description: Optional[str] = None,
        **kwargs: Any):
        """Constructs a new test case.

        Args:
            inputs: Mapping from input parameter name to input value.
            output: Optionally, an example of a desired high-quality output, or
                the output that is to be considered correct, for this test case.
            description: Optionally, a description of this test case.
            **kwargs: Additional keyword arguments to be passed to superclass
                constructor.
        """
        super().__init__(
            inputs=inputs, output=output, description=description, **kwargs)

    @classmethod
    def _from_test_suite_file_syntax(
        cls: _T_TestCase,
        params: Dict[str, Union[str, Dict[str, str]]]) -> _T_TestCase:
        """Constructs a new test case using the test suite file syntax.

        Test cases defined in a test suite file can either use the `inputs`
        keyword to specify the inputs, or all the keyword arguments will be
        treated as inputs.

        Args:
            cls: Test case class.
            params: Parameters to construct the test case.

        Returns:
            A new test case.
        
        Raises:
            TestSuiteValueError: If the test case cannot be constructed from
                the given keyword arguments.
        """
        try:
            if "inputs" not in params:
                return cls(inputs=params)
            return cls(**params)
        except TypeError as error:
            raise TestSuiteValueError(
                f"Incorrect syntax for test case: {error}."
                f"\n[yellow]Test case parameters defined in test suite file:"
                f"[/yellow] {params}") from error

    def _validate_inputs_with_signature(
        self, inputs_signature: inspect.Signature):
        """Validate that the inputs are compatible with the given signature.

        Args:
            inputs_signature: Inputs signature to validate against.

        Raises:
            TestSuiteValueError: If the inputs are not compatible with the
                signature.
        """
        try:
            inputs_signature.bind(**self.inputs)
        except TypeError as error:
            raise TestSuiteValueError(
                f"Test case inputs cannot be bound to the LLM program "
                f"signature: {error}."
                f"\n[yellow]Test case inputs:[/yellow] "
                f"{self.inputs}"
                f"\n[yellow]LLM program signature:[/yellow] "
                f"{inputs_signature}") from error


class TestCaseWithMetadata(TestCase):
    """A test case with metadata.
    
    Attributes:
        id: ID of this test case.
    """
    id: int


# Type variable for the QualityMeasure class.
_T_QualityMeasure = TypeVar("_T_QualityMeasure", bound="QualityMeasure")  # pylint: disable=invalid-name


class QualityMeasure(pydantic.BaseModel):
    """A quality measure.
    
    Attributes:
        name: Human-readable name of this quality measure.
        evaluator: Evaluator for this quality measure.  Determines whether
            this quality measure will be evaluated by running a function,
            via human inspection, or via an LLM.
        evaluation_type: The type of value produced by evaluation of this
            quality measure.
        spec: Specification of the details of how to execute this quality
            measure.
            - If evaluator is "FUNCTION", then spec should give the fully
            qualified name of the function (in the format
            "my.module:my_function").
            - If evaluator is "HUMAN", then spec should give the instructions
            or question that should be displayed to human evaluators.
            - If evaluator is "LLM", then spec should either give a dictionary
            with "model" and "prompt" fields, or the fully qualified name of an
            LLM program that implements the quality measure (in the format
            "my.module:my_function").
    """
    model_config = pydantic.ConfigDict(extra="forbid")

    name: str
    evaluator: Literal["FUNCTION", "HUMAN", "LLM"]
    evaluation_type: Literal["BINARY", "RATING_INT"]
    spec: Union[str, Dict[str, str]]

    @pydantic.field_validator("spec", mode="before")
    @classmethod
    def _extract_fully_qualified_spec_name(
        cls: _T_QualityMeasure, v: Any) -> Any:
        """Return the fully qualified name from the spec, if applicable."""
        return _extract_fully_qualified_name(v)


class QualityMeasureWithMetadata(QualityMeasure):
    """A quality measure with metadata.

    Attributes:
        id: ID of this quality measure.
    """
    id: int


# Type variable for the HparamSpec class.
_T_HparamSpec = TypeVar("_T_HparamSpec", bound="HparamSpec")  # pylint: disable=invalid-name


class HparamSpec(pydantic.BaseModel):
    """Specification of a set of hyperparameter values.

    Attributes:
        name: Name of hyperparameter.
        type: Type of hyperparameter.
        values: List of hyperparameter values.
    """
    model_config = pydantic.ConfigDict(extra="forbid", populate_by_name=True)

    name: str = pydantic.Field(alias="hparam_name")
    type: Literal["SHORT_STRING", "TEXT", "NUMBER"] = pydantic.Field(
        alias="hparam_type")
    values: Union[List[str], List[int], List[float]]

    @pydantic.model_validator(mode="before")
    @classmethod
    def _check_values_type(cls: _T_HparamSpec, data: Any) -> Any:
        """Ensure that the given values and type are aligned."""
        return wire_model.check_hparam_values_type(data)


# Type variable for the TestSuiteImportSpec class.
_T_TestSuiteImportSpec = TypeVar(  # pylint: disable=invalid-name
    "_T_TestSuiteImportSpec", bound="TestSuiteImportSpec")


class TestSuiteImportSpec(pydantic.BaseModel):
    """Specification for importing test suite components.
    
    Exactly one of test_suite_id or meta_test_suite should be non-None.
    
    Attributes:
        test_suite_id: ID of a test suite from which to import test cases.
        meta_test_suite: Meta test suite specification in the format
            "<test_suite_run_id>:<quality_measure_id>".
    """
    test_suite_id: Optional[int] = None
    meta_test_suite: Optional[str] = None

    @pydantic.model_validator(mode="after")
    def check_test_suite_id_xor_meta_test_suite(self) -> "TestSuiteImportSpec":
        """Ensure that either test_suite_id or meta_test_suite is specified.
        
        Returns:
            self
        """
        if (self.test_suite_id is None) == (self.meta_test_suite is None):
            raise ValueError(
                "Exactly one of test_suite_id or meta_test_suite must be "
                "specified.")
        return self

    @pydantic.field_validator("meta_test_suite", mode="after")
    @classmethod
    def _validate_meta_test_suite(
        cls: _T_TestSuiteImportSpec, v: Optional[str]) -> Optional[str]:
        """Validate the meta_test_suite field.
        
        Args:
            cls: The TestSuiteImportSpec class.
            v: The meta_test_suite value to be validated.
            
        Raises:
            ValueError if v does not adhere to the required format (per
            attribute description in class docstring above).
            
        Returns:
            v
        """
        if v is not None:
            try:
                if len(list(map(int, v.split(":")))) != 2:
                    raise ValueError
            except (TypeError, ValueError) as error:
                raise ValueError(
                    "meta_test_suite must be in the format "
                    "<test_suite_run_id>:<quality_measure_id>.") from error
        return v


# Type variable for the Config class.
_T_Config = TypeVar("_T_Config", bound="Config")  # pylint: disable=invalid-name


class Config(pydantic.BaseModel):
    """Config for a test suite.

    Attributes:
        id: ID of test suite.
        name: Name of test suite.
        llm_program: Fully qualified name of LLM program.
        replicas: Number of times that LLM program will be run on each pair of
            (test case, set of hyperparameters).
        parallelize: Degree of parallelism used when running the LLM program.
        autolog: Whether to automatically log select function calls that occur
            during the execution of the LLM program.
        imports: Specifications of test suites to import into the test suite
            containing this Config.
    """
    model_config = pydantic.ConfigDict(extra="forbid", populate_by_name=True)

    id: Optional[int] = None
    name: Optional[str] = None
    llm_program: str = pydantic.Field(alias="llm_program_fully_qualified_name")
    replicas: int = 1
    parallelize: int = 1
    autolog: bool = True
    imports: Optional[List[TestSuiteImportSpec]] = pydantic.Field(
        default=None,
        alias="import")

    @pydantic.model_validator(mode="after")
    def _check_id_and_name(self):
        """Ensure that either id or name is specified.
        
        Raises:
            TestSuiteValueError: If neither id nor name is specified.
        """
        if self.id is None and self.name is None:
            raise TestSuiteValueError("Either id or name must be specified.")
        return self

    @pydantic.field_validator("llm_program", mode="before")
    @classmethod
    def _extract_fully_qualified_llm_program_name(
        cls: _T_Config, v: Any) -> Any:
        """Return the fully qualified name for the program, if applicable."""
        return _extract_fully_qualified_name(v)


# A YAML file path is either a string or a path-like object.
YAMLFilePath = Union[str, os.PathLike]


def _get_components_from_file(
    path: YAMLFilePath
) -> Dict[
    str,
    Union[List[TestCase], List[QualityMeasure], List[HparamSpec], Config]]:
    """Return a dictionary of test suite components from a YAML file.
    
    Args:
        path: Path to the YAML file.
    
    Returns:
        Dictionary of test suite components, with keys:
        - test_cases: List of test cases.
        - quality_measures: List of quality measures.
        - hparam_specs: List of hyperparameter specifications.
        - config: Test suite config.
    """
    with open(path, "r", encoding="utf-8") as f:
        yaml_content = yaml.safe_load(f)

    is_list_of_dicts = (
        isinstance(yaml_content, list) and
        all(isinstance(entry, dict) for entry in yaml_content))
    if not is_list_of_dicts:
        raise TestSuiteValueError(
            "Test suite file must contain a list/sequence of "
            "dictionaries/maps.")

    test_cases = []
    quality_measures = []
    hparam_specs = []
    config = None

    for entry in yaml_content:
        for key, value in entry.items():
            try:
                if key in ["test", "test case", "test_case"]:
                    test_cases.append(
                        TestCase._from_test_suite_file_syntax(value))  # pylint: disable=protected-access
                elif key in [
                    "quality",
                    "quality measure",
                    "quality_measure",
                    "measure"]:
                    quality_measures.append(QualityMeasure(**value))
                elif key in ["hparam", "hparam spec", "hparam_spec"]:
                    hparam_specs.append(HparamSpec(**value))
                elif key in ["config", "configuration"]:
                    if config is not None:
                        raise TestSuiteValueError(
                            "Test suite file must contain at most one config "
                            "block.")
                    config = Config(**value)
                else:
                    raise TestSuiteValueError(
                        f"Invalid entry in test suite file: {key}",
                        path)

            except pydantic.ValidationError as error:
                error_messages = []
                for error_content in error.errors():
                    error_messages.append(
                        f"{error_content['msg']}: "
                        f"{', '.join(map(str, error_content['loc']))}"
                    )
                error_message = "\n".join(error_messages)
                raise TestSuiteValueError(
                    error_message,
                    path) from error

    return {
        "test_cases": test_cases,
        "quality_measures": quality_measures,
        "hparam_specs": hparam_specs,
        "config": config,
    }


# Test suite components are classes that can be added to a test suite
# (e.g., test cases, quality measures, and hyperparameter specifications)
# or files that contain test suite components (e.g., paths to YAML files).
_TestSuiteComponent = Union[
    Config, TestCase, QualityMeasure, HparamSpec, YAMLFilePath]


def get_test_suite_components(
    *args: Union[_TestSuiteComponent, List[_TestSuiteComponent]]
) -> Dict[
    str,
    Union[List[TestCase], List[QualityMeasure], List[HparamSpec], Config]]:
    """Return a dictionary of test suite components.

    Args:
        *args: One or more test cases, quality measures, hyperparameter
            specifications, paths to YAML files containing test suite
            components, or lists thereof. A config object can also be passed
            in the same way as the other components, but if more than one
            config object is found, a TestSuiteValueError is raised.

    Returns:
        Dictionary of test suite components, with keys:
        - test_cases: List of test cases.
        - quality_measures: List of quality measures.
        - hparam_specs: List of hyperparameter specifications.
        - config: Test suite config.
    
    Raises:
        TypeError: If an invalid type is passed.
        TestSuiteValueError: If more than one config block is found.
    """
    test_cases = []
    quality_measures = []
    hparam_specs = []
    config = None

    def add(arg: Union[_TestSuiteComponent, List[_TestSuiteComponent]]):
        nonlocal config
        if isinstance(arg, list):
            for item in arg:
                add(item)
        elif isinstance(arg, TestCase):
            test_cases.append(arg)
        elif isinstance(arg, QualityMeasure):
            quality_measures.append(arg)
        elif isinstance(arg, HparamSpec):
            hparam_specs.append(arg)
        elif isinstance(arg, Config):
            if config is not None:
                raise TestSuiteValueError(
                    "Test suite must contain at most one config block.")
            config = arg
        elif isinstance(arg, (str, os.PathLike)):
            add(list(_get_components_from_file(arg).values()))
        else:
            raise TypeError(
                "Invalid type. Expected TestCase, QualityMeasure, "
                f"HparamSpec, Config, or path to yaml file, but got "
                f"{type(arg)}.")

    for arg in args:
        add(arg)

    return {
        "test_cases": test_cases,
        "quality_measures": quality_measures,
        "hparam_specs": hparam_specs,
        "config": config,
    }


class TestSuite:
    """Test suite.
    
    A collection of test cases, quality measures, and hyperparameter
    specifications that can be run together.
    
    Attributes:
        config: Config for this test suite.
        test_cases: List of test cases to run.
        quality_measures: List of quality measures to compute.
        hparam_specs: List of hyperparameter specifications.
    """
    def __init__(
        self,
        id_or_name: Union[int, str],
        llm_program: Union[Callable, str]):
        """Create a test suite.
        
        Args:
            id_or_name: ID or name of the test suite.
            llm_program: LLM program to test. Either a Python object or a
                string containing the fully qualified name of the Python
                object. The Python object can be either a Python function or
                LangChain chain. If a string is passed, it must be in the
                format:
                `<fully qualified module name>:<fully qualified object name>`.
        """
        id, name = None, None  # pylint: disable=redefined-builtin
        if isinstance(id_or_name, int):
            id = id_or_name
        elif isinstance(id_or_name, str):
            name = id_or_name
        else:
            raise TypeError(
                "Invalid type for id_or_name. Expected int or str, "
                f"but got {type(id_or_name)}.")

        self.config = Config(
            id=id,
            name=name,
            llm_program=llm_program)
        self.test_cases: List[TestCase] = []
        self.quality_measures: List[QualityMeasure] = []
        self.hparam_specs: List[HparamSpec] = []

        self._auth_session = auth_session.get_auth_session()

    def add(
        self,
        *args: Union[_TestSuiteComponent, List[_TestSuiteComponent]]):
        """Add test cases, quality measures, or hyperparameter specifications.

        Args:
            *args: One or more test cases, quality measures, hyperparameter
                specifications, or path to a YAML file containing test suite
                components to add to the test suite. If a list is passed,
                each item in the list is added to the test suite. Any Config
                objects passed are ignored.
        
        Raises:
            TypeError: If an invalid type is passed.
        """
        components = get_test_suite_components(*args)
        self.test_cases.extend(components["test_cases"])
        self.quality_measures.extend(components["quality_measures"])
        self.hparam_specs.extend(components["hparam_specs"])

    def import_test_cases(
        self,
        test_suite_id: Optional[int] = None,
        *,
        meta_test_suite: Optional[str] = None
    ):
        """Import test cases from another test suite.

        Stores the details of the import in this test suite's config. Only when
        this test suite is run will the test cases from the specified test suite
        be retrieved and included in the test suite run.
        
        Exactly one of test_suite_id or meta_test_suite must be non-None.
        
        Args:
            test_suite_id: ID of a test suite from which to import test cases.
            meta_test_suite: Meta test suite specification in the format
                "<test_suite_run_id>:<quality_measure_id>".
        """
        new_import_spec = TestSuiteImportSpec(
            test_suite_id=test_suite_id,
            meta_test_suite=meta_test_suite)

        if self.config.imports is None:
            self.config.imports = [new_import_spec]
        else:
            self.config.imports.append(new_import_spec)


    def _validate(self):
        """Validate the test suite.
        
        Perform the following checks to ensure that the test suite is valid:
        - LLM program is callable.
        - Test cases inputs can be bound to the LLM program signature.

        Raises:
            TestSuiteValueError: If the test suite is invalid.
        """
        # Check that the LLM program is callable.
        llm_program = self.config.llm_program
        try:
            llm_program_callable = LazyCallable(
                llm_program).get_callable()
        except Exception as error:
            # It is not trival identify the relevant Exception subtypes that
            # should be considered actual validation errors, so we catch all
            # exceptions and raise a generic validation error with the
            # exception message appended.
            raise TestSuiteValueError(
                f"LLM program {llm_program} is not callable. "
                f"{error}") from error

        # Check that the test cases inputs can be bound to the LLM program
        # signature.
        # TODO(#323): Include validation for LangChain objects.
        if inspect.isfunction(llm_program_callable):
            llm_program_signature = inspect.signature(llm_program_callable)
            for test_case in self.test_cases:
                test_case._validate_inputs_with_signature(llm_program_signature)  # pylint: disable=protected-access

        # TODO: https://github.com/inductor-hq/saas/issues/324 - Remove any
        # duplicate test suite components.

    def run(
        self,
        *,
        replicas: Optional[int] = None,
        parallelize: Optional[int] = None):
        """Run the test suite.

        Args:
            replicas: Number of replicated executions to perform for each
                (test case, unique set of hyperparameter values) pair. Defaults
                to 1.
            parallelize: Number of LLM program executions to run in parallel.
                Defaults to 1.
        """
        auth_access_token = self._auth_session.access_token

        if replicas is None and parallelize is None:
            self._run(auth_access_token)
        else:
            # Test suites are run using the settings in their own config. In
            # order to functionally modify the config for only this run, we
            # create a shallow copy of the test suite and its config.
            test_suite_to_run = copy.copy(self)
            test_suite_to_run.config = copy.copy(self.config)
            if replicas is not None:
                test_suite_to_run.config.replicas = replicas
            if parallelize is not None:
                test_suite_to_run.config.parallelize = parallelize

            test_suite_to_run._run(auth_access_token)  # pylint: disable=protected-access

    def _run(
        self,
        auth_access_token: str,
        *,
        prompt_open_results: bool = False):
        """Run the test suite using its current config.
        
        Args:
            auth_access_token: Access token for authentication.
            prompt_open_results: Whether to prompt the user to open the
                results in the browser.
        """
        self._validate()
        # Import here to avoid circular imports.
        from inductor.execution.test_suite_execution import execute_test_suite  # pylint: disable=import-outside-toplevel
        execute_test_suite(
            test_suite=self,
            auth_access_token=auth_access_token,
            prompt_open_results=prompt_open_results)

    def _get_run_request(self) -> wire_model.CreateTestSuiteRunRequest:
        """Return a CreateTestSuiteRunRequest for this test suite.
        
        Returns:
            A CreateTestSuiteRunRequest for this test suite.
        """
        return wire_model.CreateTestSuiteRunRequest(
            test_suite_id_or_name=(
                self.config.id or self.config.name),
            test_cases=[
                test_case.model_dump()
                for test_case in self.test_cases],
            quality_measures=[
                quality_measure.model_dump()
                for quality_measure in self.quality_measures],
            hparam_specs=[
                hparam_spec.model_dump()
                for hparam_spec in self.hparam_specs],
            llm_program_details=LazyCallable(
                self.config.llm_program).get_program_details(),
            replicas=self.config.replicas,
            parallelize=self.config.parallelize,
            test_suite_import_specs=[
                import_spec.model_dump()
                for import_spec in self.config.imports or []] or None
        )


def get_test_suites_via_cli_args(cli_args: Dict[str, Any]) -> List[TestSuite]:
    """Return a list of test suites.

    Parse the relevant arguments from the given command line arguments.
    Use a combination of the command line arguments and the test suite file
    contents (where the test suite file is specified by the required
    `test_suite_file_paths` command line argument) to construct the test
    suites. The config for each test suite is constructed based on the
    following hierarchy of configuration sources:
    1. Command line arguments.
    2. Test suite file arguments.
    3. Default arguments.
    In the case of a conflict, the lower-numbered item in the list takes
    precedence.
    
    Args:
        cli_args: Dictionary of command line arguments. Note that not all the
            command line arguments are used in this function.

    Returns:
        A list of `TestSuite` objects.

    Raises:
        TestSuiteValueError: If any test suite is invalid.
    """
    test_suite_file_paths = cli_args.get("test_suite_file_paths")
    if test_suite_file_paths is None:
        raise ValueError("`test_suite_file_paths` must be defined.")

    test_suites = []
    for test_suite_file_path in test_suite_file_paths:
        try:
            # Get the test suite components from the test suite file.
            test_suite_components = _get_components_from_file(
                test_suite_file_path)
            test_suite_config: Optional[Config] = test_suite_components[
                "config"]
            if test_suite_config is None:
                raise TestSuiteValueError(
                    "Test suite file must contain a config.")

            # Update the test suite config with the command line arguments.
            for key, value in cli_args.items():
                if key in Config.model_fields and value is not None:
                    setattr(test_suite_config, key, value)

            # Create the test suite.
            test_suite = TestSuite(
                id_or_name=test_suite_config.id or test_suite_config.name,
                llm_program=test_suite_config.llm_program)
            test_suite.config = test_suite_config
            test_suite.add(list(test_suite_components.values()))
            test_suites.append(test_suite)

        except TestSuiteValueError as error:
            # Add details about the source of the error.
            error.path = test_suite_file_path
            raise error

    return test_suites


def components_with_ids(
    components: List[Union[TestCase, QualityMeasure]],
    ids: List[int],
    component_with_id_type: Union[
        Type[TestCaseWithMetadata], Type[QualityMeasureWithMetadata]]
) -> List[Union[TestCaseWithMetadata, QualityMeasureWithMetadata]]:
    """Returns components with their IDs.

    Args:
        components: List of components.
        ids: List of IDs in the same order as the components. (i.e.,
            `ids[i]` is the ID of `components[i]`.) The length of `ids` must
            match the length of `components`.
        component_with_id_type: Class that extends the component type with
            an `id` attribute.

    Returns:
        List of components with their IDs.
    
    Raises:
        ValueError: If the number of components does not match the number of
            IDs.
    """
    if len(components) != len(ids):
        raise ValueError(
            "Number of components does not match number of IDs.")
    return [
        component_with_id_type(
            id=component_id,
            **component.model_dump())
        for component_id, component in zip(ids, components)
    ]


def in_google_colab() -> bool:
    """Return True if currently running in Google Colab, and False otherwise."""
    try:
        import google.colab  # pylint: disable=import-outside-toplevel,unused-import
        return True
    except ImportError:
        return False


def get_module_qualname(f: Callable) -> str:
    """Return the fully qualified name of the module in which f is defined.

    Args:
        f: A function, class, or method.

    Returns:
        The fully qualified name of the module in which f is defined. If f is
        defined in the __main__ module, then the name of the file containing f
        (without its ".py" extension) is returned as the fully qualified module
        name. If f is defined in a Google Colab notebook, then "google_colab" is
        returned as the fully qualified module name.

    Raises:
        RuntimeError if f is defined in the __main__ module, the current
        environment is not Google Colab, and the name of the file
        containing f does not end with ".py".
    """
    qualname = f.__module__
    if qualname == "__main__":
        # Special case for Google Colab.
        if in_google_colab():
            return "google_colab"

        qualname, ext = os.path.splitext(
            os.path.basename(f.__globals__["__file__"]))
        if ext != ".py":
            raise RuntimeError(
                f"f ({f.__qualname__}) is defined in the __main__ module but "
                f"is contained in a file ({f.__globals__['__file__']}) that "
                "does not have extension '.py'.")
    return qualname


def deepcopy_or_str(value: Any) -> Any:
    """Return a deep copy of the given value or a string representation of it.

    If the value is not deep copyable or not JSON serializable, it will be
    converted to a string. Otherwise, a deep copy of the value will be returned.

    Args:
        value: Value to deep copy or convert to a string.
    
    Returns:
        Deep copy of the given value, if applicable, otherwise the given value
        converted to a string.
    """
    try:
        json.dumps(value)
        value = copy.deepcopy(value)
    except (TypeError, OverflowError, RecursionError, AttributeError):
        value = str(value)
    return value


class LazyCallable:
    """Container for a function or LangChain object that is lazily imported.
    
    Represents a function or LangChain object that is provided by the user as
    part of a test suite. Objects of this class are callable, and calling them
    will import the required module if necessary to call the object.

    Attributes:
        fully_qualified_name (str): Fully qualified name of the callable
            object, in the format "<fully qualified module name>:<fully
            qualified object name>". (e.g. "my.module:my_function")
        path_to_module_dir (str): Path to the directory that contains the
            module that contains the callable object.
        input_signature (Dict[str, Optional[str]]): Inputs signature of the
            callable object. Inputs signature is a map between input parameter
            names to strings indicating the corresponding parameter types (or to
            None for input parameters that do not have type annotations).
    """
    def __init__(
        self,
        callable_or_fully_qualified_name: Union[Callable, str],
        path_to_module_dir: str = os.getcwd()):
        """Initialize a LazyCallable object.

        Args:
            callable_or_fully_qualified_name: Callable object or fully
                qualified name of the callable object, in the format
                "<fully qualified module name>:<fully qualified object name>".
                (e.g. "my.module:my_function")
            path_to_module_dir: Path to the directory that contains the module
                that contains the callable object. If not provided, the
                current working directory is used. This argument is ignored if
                `callable_or_fully_qualified_name` is a callable object.
        """
        if isinstance(callable_or_fully_qualified_name, Callable):
            callable_object = callable_or_fully_qualified_name
            module_qualname = get_module_qualname(callable_object)
            object_qualname = callable_object.__qualname__

            self.fully_qualified_name = f"{module_qualname}:{object_qualname}"
            self.path_to_module_dir = os.path.dirname(
                inspect.getfile(callable_object))
        elif isinstance(callable_or_fully_qualified_name, str):
            self.fully_qualified_name = callable_or_fully_qualified_name
            self.path_to_module_dir = path_to_module_dir

            callable_object = self.get_callable()
        else:
            raise ValueError(
                "callable_or_fully_qualified_name must be a callable or a "
                "string.")

        if callable_object.__class__.__name__ == "LLMChain":
            self._program_type = "LANGCHAIN"
        elif inspect.isfunction(callable_object):
            self._program_type = "FUNCTION"
        else:
            raise ValueError(
                f"Object {self.fully_qualified_name} is not a function or "
                "LangChain object.")

        self.inputs_signature = self._get_inputs_signature(callable_object)

    def get_program_details(self) -> wire_model.ProgramDetails:
        """Return a `wire_model.ProgramDetails` object.
        
        Returns:
            `wire_model.ProgramDetails` object.
        """
        return wire_model.ProgramDetails(
            fully_qualified_name=self.fully_qualified_name,
            program_type=self._program_type,
            inputs_signature=self.inputs_signature)

    def _import_module(self, module_qualname: str) -> types.ModuleType:
        """Import the module given by module_qualname.

        Args:
            module_qualname: Fully qualified name of the module to import.

        Returns:
            The imported module.
        """
        orig_sys_path = sys.path.copy()
        sys.path[0] = self.path_to_module_dir
        module = importlib.import_module(module_qualname)
        sys.path = orig_sys_path
        return module

    def get_callable(self) -> Callable:
        """Import the callable object and return it.
        
        Returns:
            Callable object.
        """
        module_qualname, object_qualname = (
            self.fully_qualified_name.split(":"))
        module_qualname = module_qualname.replace(
            "/", ".")
        # Note: we do not use removesuffix() here in order to enable
        # compatibility with Python 3.8.
        if module_qualname.endswith(".py"):
            module_qualname = module_qualname[:-len(".py")]

        module_is_main = (module_qualname == "google_colab") or (
            # To prevent re-importing the __main__ module under the name
            # given by its filename (in order to prevent inadvertently
            # re-executing the __main__ module's contents).
            "__main__" in sys.modules and
            hasattr(sys.modules["__main__"], "__file__") and
            module_qualname == os.path.splitext(
                os.path.basename(sys.modules["__main__"].__file__))[0])  # pylint: disable=no-member
        if module_is_main:
            module_qualname = "__main__"

        module = self._import_module(module_qualname)
        callable_object = module
        for name in object_qualname.split("."):
            callable_object = getattr(callable_object, name)
        return callable_object

    def __call__(self, *args, **kwargs) -> Any:
        """Call the callable object.
        
        Args:
            *args: Positional arguments to pass to the callable object.
            **kwargs: Keyword arguments to pass to the callable object.
        
        Returns:
            The return value of the callable object.
        """
        callable_object = self.get_callable()
        if self._program_type == "LANGCHAIN":
            return callable_object.run(*args, **kwargs)
        else:
            if asyncio.iscoroutinefunction(callable_object):
                return asyncio.run(callable_object(*args, **kwargs))
            return callable_object(*args, **kwargs)

    def _get_inputs_signature(
        self,
        callable_object: Optional[Callable] = None
        ) -> Dict[str, Optional[str]]:
        """Return the inputs signature of the callable object.
        
        Inputs signature is a map between input parameter names to
        strings indicating the corresponding parameter types (or to None for
        input parameters that do not have type annotations).

        Args:
            callable_object: Callable object. If not provided, the callable
                object will be imported.
        
        Returns:
            Inputs signature of the callable object.
        """
        if callable_object is None:
            callable_object = self.get_callable()
        if self._program_type == "LANGCHAIN":
            return {key: None for key in callable_object.input_keys}
        else:
            signature = inspect.signature(callable_object)
            return {
                name: (
                    str(param.annotation)
                    if param.annotation != inspect._empty  # pylint: disable=protected-access
                    else None)
                for name, param in signature.parameters.items()
            }


class LiveExecution:
    """Object linking to an LLM program post execution after it has returned."""

    def __init__(
        self,
        llm_program_execution_auth_session: Optional[auth_session.AuthSession]
    ):
        """Initialize a LlmProgramExecution object.

        Args:
            llm_program_execution_auth_session: AuthSession object representing
                the session for this LLM program execution.
        """
        # Initialize identifier for the LLM program execution as None.
        # This will be set later asynchronously by `_set_id`.
        self._execution_id = None
        # Create list of LoggedValue objects. Used to store logged values
        # before an execution_id is set.
        self._logged_values = []

        self._auth_session = llm_program_execution_auth_session

    def _set_id(self, execution_id: int):
        """Sets the LLM program execution id field and flushes logged values.

        Args:
            execution_id: ID of the LLM program execution.
        """
        self._execution_id = execution_id
        # Flush any values logged before the id was set
        if self._logged_values:
            self._send_log_request(self._logged_values)
            self._logged_values = []

    def _send_log_request(
        self,
        logged_values_list: List[wire_model.LoggedValue]
    ):
        """Sends the log information to the client API.

        Log information is only sent if self._auth_session is not None.
        This function should only be called if self._execution_id is not None.

        Args:
            logged_values_list: List of LoggedValue objects.
        """
        if self._auth_session is not None:
            request = wire_model.LogValuesForExecutionRequest(
                execution_id=self._execution_id,
                logged_values=logged_values_list
            )
            auth_access_token = self._auth_session.access_token
            backend_client.log_values_for_execution(request, auth_access_token)

    def log(self, value: Any, *, name: Optional[str] = None):
        """Log a value and associate it with the LLM program execution.

        Args:
            value: The value to be logged.
            name: An optional human-readable name for the logged value.
        """
        logged_value = wire_model.LoggedValue(
            value=deepcopy_or_str(value),
            description=name,
            after_complete=True)
        if self._execution_id is not None:
            self._send_log_request([logged_value])
        else:
            self._logged_values.append(logged_value)



# Copyright 2023 Inductor, Inc.
"""Functionality for general LLM program execution tasks."""

import contextlib
import contextvars
import io
import itertools
import sys
import traceback
from typing import Any, Callable, Dict, Iterator, List, Optional, TextIO, Tuple, TypeVar, Union

from inductor.backend_client import wire_model
from inductor.data_model import data_model


# The following module-private variables are used to store information
# about the current LLM program execution. Context variables are used to
# enable running multiple LLM program executions in parallel using
# multiple threads. While an LLM program itself may use multiple threads,
# those nested threads can not interact with the LLM program's execution
# (e.g., by calling inductor.log()).
#
# Whether the logger decorator (inductor.logger) is enabled. This is set to
# False when running tests to prevent the logger from sending duplicate data to
# the backend, in the case that the LLM program being tested uses the logger
# decorator.
logger_decorator_enabled = True
# Stores a list of logged values for the current LLM program execution.
logged_values = contextvars.ContextVar("logged_values", default=None)
# Dictionary of hyperparameter values for the current LLM program execution.
hparams = {}


@contextlib.contextmanager
def set_hparams(
    hparams_to_set: Dict[str, Union[str, int, float]],
    primary_execution: bool = True):
    """Set execution.hparams to the given hparams within this context manager.
    
    Set the execution.hparams module-level variable to the given
    hyperparameters. On exit, set `execution.hparams` to its original value
    (i.e., the value it had before entering this context manager).

    Args:
        hparams_to_set: A dictionary mapping hyperparameter names to values.
        primary_execution: Whether the current LLM program execution is the
            primary execution. (This flag is relevant to live executions
            which may or may not be a primary execution. See
            `inductor.execution.live_execution._primary_execution` for
            details. A test suite run is always considered a primary
            execution.) If False, the current hyperparameters will not be
            overwritten.
    """
    if primary_execution:
        global hparams
        original_hparams = hparams
        try:
            hparams = hparams_to_set
            yield
        finally:
            hparams = original_hparams
    else:
        yield


def hparam(
    name: str,
    default_value: Union[str, int, float]) -> Union[str, int, float]:
    """Return the value of the hyperparameter having the given name.

    Args:
        name: Name of hyperparameter value to be returned.
        default_value: Value that will be returned if a value has not
            been specified for the given name.
    """
    if name in hparams:
        return hparams[name]
    else:
        return default_value


def _log(
    value: Any, *, after_complete: bool, description: Optional[str] = None):
    """Log a value and associate it with the current LLM program execution.

    Args:
        value: The value to be logged.
        after_complete: Whether the value was logged after the LLM
            program execution completed.
        description: An optional human-readable description of the logged
            value.
    """
    logged_values_list = logged_values.get()
    if logged_values_list is None:
        # We can not distinguish between the below two cases described in the
        # printed message, so we print the same message for both cases.
        print(
            f"[WARNING] inductor.log call failed with "
            f"{{\"value\": {value}, \"name\": {description}}}.\n"
            "Cannot call inductor.log outside of a function "
            "decorated with @inductor.logger, unless you are running "
            "a test suite. This log call will be ignored. "
            "Also note that invoking inductor.log from a thread different "
            "from the one that initialized the logger (via the decorator or "
            "the CLI tool) is currently unsupported. If you require support "
            "for this, please contact Inductor support to submit a feature "
            "request.")
    else:
        logged_values_list.append(
            wire_model.LoggedValue(
                value=data_model.deepcopy_or_str(value),
                description=description,
                after_complete=after_complete))


def log(value: Any, *, name: Optional[str] = None):
    """Log a value and associate it with the current LLM program execution.

    Args:
        value: The value to be logged.
        name: An optional human-readable name for the logged value.
    
    Raises:
        RuntimeError: If the LLM program execution was not initiated via the
            Inductor CLI, and the LLM program is not decorated with
            @inductor.logger.
    """
    _log(value, description=name, after_complete=False)


def get_hparams_combinations(
    hparam_specs: Optional[List[data_model.HparamSpec]] = None
    ) -> List[Dict[str, Union[str, int, float]]]:
    """Get all combinations of hyperparameters.

    Given a list of hyperparameters and their possible values, return a list of
    dictionaries, where each dictionary represents a unique combination of
    hyperparameters.

    For example, if the given hyperparameters are:
    [
        data_model.HparamSpec(
            hparam_name="a",
            values=[1, 2],
        ),
        data_model.HparamSpec(
            hparam_name="b",
            values=[3, 4],
        ),
    ]
    then the returned list will be:
    [
        {"a": 1, "b": 3},
        {"a": 1, "b": 4},
        {"a": 2, "b": 3},
        {"a": 2, "b": 4},
    ]

    Args:
        hparam_specs: List of hyperparameters specs, where all hyperparameter
            specs have distinct names.

    Returns:
        A list of dictionaries, where each dictionary represents a unique
            combination of hyperparameters.

    Raises:
        ValueError: If hyperparameter names are not distinct.
    """
    if hparam_specs is None:
        return [{}]

    # Convert list of HparamSpec to dictionary
    hparams_dict = {
        hparam.name: hparam.values for hparam in hparam_specs}

    # Ensure that all hyperparameter names in hparam_specs are distinct.
    if len(hparams_dict) != len(hparam_specs):
        raise ValueError(
            "Hyperparameter names in hparam_specs must be distinct.")

    keys = list(hparams_dict.keys())
    value_lists = [hparams_dict[key] for key in keys]

    # Generate combinations.
    value_combinations = list(itertools.product(*value_lists))

    # Convert to dictionaries.
    hparam_combinations = []
    for value_combination in value_combinations:
        hparam_combinations.append(dict(zip(keys, value_combination)))

    return hparam_combinations


# Type variable for IteratorWrapper objects.
_T_IteratorWrapper = TypeVar(  # pylint: disable=invalid-name
    "_T_IteratorWrapper", bound="IteratorWrapper")


class IteratorWrapper:
    """Iterator wrapper.

    Wraps an iterator and captures values yielded by the iterator. When the
    iterator is exhausted or an error (raised by the underlying iterator)
    occurs during iteration, the wrapper calls a given `stop_signal_handler`
    function. The IteratorWrapper instance is passed to the
    `stop_signal_handler` function as the keyword argument `output`. The stop
    signal handler function can also be passed additional keyword arguments
    via the `stop_signal_handler_kwargs` argument.
    """
    def __init__(
        self,
        iterator: Iterator,
        *,
        stop_signal_handler: Callable,
        stop_signal_handler_kwargs: Optional[Dict[str, Any]] = None,
        iterator_wrapper_error_message: str):
        """Create an IteratorWrapper.
        
        Args:
            iterator: The iterator to wrap.
            stop_signal_handler: Function to call when the iterator is
                exhausted or an error (raised by the underlying iterator)
                occurs during iteration. This function is required to accept
                the keyword argument `output`, which will be this
                `IteratorWrapper` instance (which is of type
                `IteratorWrapper`). If `stop_signal_handler_kwargs` is not
                None, any keyword arguments passed via 
                `stop_signal_handler_kwargs` will also be passed to
                `stop_signal_handler` when it is called.
            stop_signal_handler_kwargs: Keyword arguments to pass to
                `stop_signal_handler` when it is called. If
                `stop_signal_handler_kwargs` includes the key "output",
                an error will be raised as this key is reserved.
            iterator_wrapper_error_message: Error message to print if an error
                occurs within this `IteratorWrapper` instance as part of
                capturing values yielded by the iterator, or if an error
                occurs within the `stop_signal_handler` function. The actual
                error that occurred will be printed after this message.
        
        Raises:
            ValueError: If `stop_signal_handler_kwargs` includes the key
                "output".
        """
        self._iterator = iterator

        self._completed_values = []
        self._skip_stop_signal_handler = False
        self._stop_signal_occurred = False
        self._iteration_error = None

        self._stop_signal_handler = stop_signal_handler
        if stop_signal_handler_kwargs is None:
            stop_signal_handler_kwargs = {}
        self._stop_signal_handler_kwargs = stop_signal_handler_kwargs.copy()
        if "output" in self._stop_signal_handler_kwargs:
            raise ValueError(
                "The key 'output' is reserved and cannot be "
                "passed in `stop_signal_handler_kwargs`.")
        self._stop_signal_handler_kwargs["output"] = self

        self._iterator_wrapper_error_message = iterator_wrapper_error_message

    def __iter__(self) -> _T_IteratorWrapper:
        return self

    def __next__(self) -> Any:
        """Get the next value from the iterator.

        Before returning the next value from the iterator, append a deep copy
        of the value or a string representation of the value to the list of
        completed values. (See `data_model.deepcopy_or_str` for details.)
        If the iterator is exhausted or an error occurs during iteration, call
        the `stop_signal_handler` function. Print the
        `iterator_wrapper_error_message` if an error occurs within this
        `IteratorWrapper` instance as part of capturing values yielded by the
        iterator, or if an error occurs within the `stop_signal_handler`
        function.

        Returns:
            The next value from the iterator.

        Raises:
            StopIteration: If the iterator is exhausted.
        """
        try:
            value = next(self._iterator)
        except StopIteration as stop_signal:
            self._stop_signal_occurred = True
            raise stop_signal
        except Exception as error:  # pylint: disable=broad-except
            self._iteration_error = error
            raise error
        finally:
            if not self._skip_stop_signal_handler:
                try:
                    if (self._stop_signal_occurred or
                        self._iteration_error is not None):
                        if all(
                            isinstance(value, str)
                            for value in self._completed_values
                        ):
                            self._completed_values = "".join(
                                self._completed_values)

                        self._stop_signal_handler(
                            **self._stop_signal_handler_kwargs)

                        self._skip_stop_signal_handler = True
                    else:
                        self._completed_values.append(
                            data_model.deepcopy_or_str(value))  # pylint: disable=used-before-assignment

                except Exception as wrapper_error:  # pylint: disable=broad-except
                    self._skip_stop_signal_handler = True
                    traceback.print_exc()
                    print(self._iterator_wrapper_error_message)
                    print(wrapper_error)

        return value

    def __getattr__(self, name: str) -> Any:
        """Forward unknown attributes to the iterator.
        
        This allows the iterator wrapper to be used as if it were the
        underlying iterator.

        Args:
            name: Name of attribute to get.

        Returns:
            The requested attribute.
        """
        return getattr(self._iterator, name)

    def _get_completed_values(self) -> Union[List[Any], str]:
        """Get the values that were yielded by the iterator.

        If all of the values are strings, the values are joined into a single
        string.

        Returns:
            The values that were yielded by the iterator.
        """
        completed_values = self._completed_values
        if not isinstance(completed_values, str) and all(
            isinstance(value, str)
            for value in completed_values
        ):
            completed_values = "".join(completed_values)
        return completed_values


@contextlib.contextmanager
def capture_stdout_stderr(
    suppress: bool = False) -> Tuple[io.StringIO, io.StringIO]:
    """Capture stdout and stderr.
    
    On exit, restore the original stdout and stderr and close the yielded
    StringIO buffers (i.e., the yielded buffers' contents will be discarded
    when context manager exits).
    
    Args:
        suppress: Whether to suppress stdout and stderr. If True, the
            contents of stdout and stderr will be suppressed after being
            captured. If False, stdout and stderr will behave as normal,
            but their contents will still be captured.

    Yields:
        A tuple of streams used to capture stdout and stderr.
    """
    class Tee(io.StringIO):
        """A StringIO buffer that optionally writes to a file in addition to
        capturing the written string."""
        def __init__(self, file: Optional[TextIO]):
            """Override the constructor to store the file to which to write."""
            self.file = file
            super().__init__()

        def write(self, s: str):
            """Override the write method to write to the file (as merited)
            in addition to capturing the written string."""
            if self.file is not None:
                self.file.write(s)
            return super().write(s)

    stdout_capture = Tee(
        sys.stdout if not suppress else None)
    stderr_capture = Tee(
        sys.stderr if not suppress else None)

    # Save the original stdout and stderr.
    original_stdout = sys.stdout
    original_stderr = sys.stderr
    # Redirect stdout and stderr to the Tee objects.
    sys.stdout = stdout_capture
    sys.stderr = stderr_capture
    try:
        yield (stdout_capture, stderr_capture)
    finally:
        # Restore the original stdout and stderr.
        sys.stdout = original_stdout
        sys.stderr = original_stderr
        # Close the StringIO buffers.
        stdout_capture.close()
        stderr_capture.close()


@contextlib.contextmanager
def capture_logged_values():
    """Capture values logged via log() calls.
    
    If logging has not already been initialized, initialize logging by setting
    the logged values context variable (`_logged_values`) to an empty list,
    and, on exit, set `_logged_values` to `None`.
    If logging has already been initialized, do nothing.
    In either case, yield the list of logged values.

    The purpose of this context manager is to manage the state of the
    logged values context variable, which should only be initialized
    once per LLM program execution.

    Yields:
        The list of logged values.
    """
    init_logged_values = logged_values.get()
    initializing_logged_values = init_logged_values is None
    try:
        if initializing_logged_values:
            logged_values.set([])
        yield logged_values.get()
    finally:
        if initializing_logged_values:
            logged_values.set(None)



# Copyright 2023 Inductor, Inc.
"""Collection of Questions / User Prompts for the Inductor CLI.

This module is intended to include questions that require non-trivial
validation or other complex logic. It does not need to include simple
questions.
"""

import pathlib
import re
from typing import Any, Dict, List, Optional
import inquirer
import rich

from inductor.data_model import data_model


def get_new_file_path(
    default_path: str,
    accepted_extensions: List[str],
    param_name: str,
    prefix: Optional[str] = None
) -> pathlib.Path:
    """Prompt user for a path to a new file.

    Prompt the user to enter a path for a new file. Validate the path. If the
    path is valid, return it. If the path is invalid, continue prompting the
    user until a valid path is entered.

    Args:
        default_path: Default path to display to the user.
        accepted_extensions: List of accepted extensions for the path. The
            extensions should include the leading period (e.g. ".py").
        param_name: Name of the param to display to the user.
        prefix: Prefix to display to the user before the prompt. Displayed
            inside brackets.

    Returns:
        Path to the new file.
    """
    default_path: pathlib.Path = pathlib.Path(default_path)
    path = pathlib.Path(default_path)
    i = 1
    while path.exists():
        path = pathlib.Path(f"{default_path.stem}_{i}{default_path.suffix}")
        i += 1
        if i > 100:
            # If we've tried 100 times, something is wrong.
            # Use the initial default path.
            path = default_path
            break

    # TODO (https://github.com/inductor-hq/saas/issues/18): Add autocomplete
    # for paths.
    def validate_path(_: Dict[str, Any], current: str) -> bool:
        """Validate a given path.

        A path is valid if it:
        1. Has an accepted extension.
        2. Does not already exist.
        3. Can be touched.

        Given a path to a file, validate it. If the path is valid and gets
        created, delete it.

        Args:
            _: Dictionary containing the answers from previous questions
                (not including the current question). This will be an empty
                dictionary because the `file_path` question is not part of a
                collection of questions.
            current: The current answer, which is a path to validate.

        Raises:
            inquirer.errors.ValidationError: If the path is invalid.
        
        Returns:
            True if the path is valid, False otherwise.
        """
        try:
            path = pathlib.Path(current).resolve()
            if path.suffix not in accepted_extensions:
                # NOTE: This will also catch empty paths.
                raise inquirer.errors.ValidationError(
                    "",
                    reason=("File requires an extension in "
                            f"{accepted_extensions}."))

            # Permissions: rw-rw-rw- (0o666)
            path.touch(mode=0o666, exist_ok=False)
            if path.exists():
                path.unlink()
        except FileExistsError:
            raise inquirer.errors.ValidationError(  # pylint: disable=raise-missing-from
                "", reason="File already exists.")
        except PermissionError:
            raise inquirer.errors.ValidationError(  # pylint: disable=raise-missing-from
                "", reason="Missing permissions to create this file.")
        except FileNotFoundError:
            # The path specified by current does not exist.
            raise inquirer.errors.ValidationError(  # pylint: disable=raise-missing-from
                "", reason="Invalid path.")
        return True

    info = (f"Enter the file path for the new {param_name}, ensuring it has an "
            f"extension in {accepted_extensions} and does not already "
            "exist.")
    if prefix is not None:
        info = f"[[yellow]{prefix}[/yellow]] {info}"
    rich.print(f"\n{info}")
    new_file_path = inquirer.text(
        message=f"{param_name.capitalize()} file path",
        default=path,
        validate=validate_path)
    return pathlib.Path(new_file_path).resolve()


def get_test_suite_name(
    default_name: Optional[str] = None,
    param_name: Optional[str] = "test suite",
    prefix: Optional[str] = None
) -> str:
    """Prompt user for a test suite name.

    Prompt the user to enter a test suite name. Validate the name (see
    `validate_test_suite_name` function below). If the name is valid,
    return it. If the name is invalid, continue prompting the user until a
    valid name is entered.

    Args:
        default_name: Default test suite name to display to the user.
        param_name: Name of the test suite to display to the user.
        prefix: Prefix to display to the user before the prompt.  Displayed
            inside brackets.

    Returns:
        Test suite name.
    """
    def validate_test_suite_name(_: Dict[str, Any], current: str) -> bool:
        """Validate test suite name.

        A test suite name is valid if it:
        1. Is non-empty.
        2. Includes only alphanumeric characters, underscores, or dashes.

        Args:
            _: Dictionary containing the answers from previous questions
                (not including the current question). This will be an empty
                dictionary because the `test_suite_name` question is not
                part of a collection of questions.
            current: The current answer, which is a test suite name to
                validate.

        Returns:
            True if the test suite name is valid, False otherwise.
        
        Raises:
            inquirer.errors.ValidationError: If the test suite name is
                invalid.
        """
        name = current

        if not re.match(r"^[a-zA-Z0-9_-]+$", name):
            raise inquirer.errors.ValidationError(
                "", reason="Invalid test suite name.")
        return True

    info = (f"Enter a name for the new {param_name}, ensuring that it "
            "is non-empty and includes only alphanumeric characters, "
            "underscores, or dashes. Your test suites must have unique names.")
    if prefix is not None:
        info = f"[[yellow]{prefix}[/yellow]] {info}"
    rich.print(f"\n{info}")

    return inquirer.text(
        message=f"{param_name.capitalize()} name",
        default=default_name,
        validate=validate_test_suite_name)


def get_test_suite_description(prefix: Optional[str] = None) -> Optional[str]:
    """Prompt user for a test suite description.

    Args:
        prefix: Prefix to display to the user before the prompt. Displayed
            inside brackets.
    
    Returns:
        Test suite description, or None if the user skipped this question.
    """
    info = ("Optionally enter a description for the new test suite. Press "
            "enter to skip.")
    if prefix is not None:
        info = f"[[yellow]{prefix}[/yellow]] {info}"
    rich.print(f"\n{info}")

    description = inquirer.text(message="Test suite description")
    if not description:
        return None
    return description


def get_llm_program_fully_qualified_name(
    prefix: Optional[str] = None
) -> str:
    """Prompt user for a LLM program fully qualified name.

    Prompt the user to enter a LLM program fully qualified name. Validate the
    name (see `validate_llm_program` function below). If the name is valid,
    return it. If the name is invalid, continue prompting the user until a
    valid name is entered.

    Args:
        prefix: Prefix to display to the user before the prompt. Displayed
            inside brackets.

    Returns:
        LLM program fully qualified name.
    """
    def validate_llm_program(_: Dict[str, Any], current: str) -> bool:
        """Validate LLM program fully qualified name.

        A LLM program fully qualified name is valid if it:
        1. Is in the format <fully qualified module name>:<fully qualified
            object name>.
        2. Can instantiate a `data_model.LazyCallable` object.
        3. The instantiated object can import the underlying LLM program
            and get its parameter keys with `get_parameter_keys`.

        Args:
            _: Dictionary containing the answers from previous questions
                (not including the current question). This will be an empty
                dictionary because the `llm_program_fully_qualified_name`
                question is not part of a collection of questions.
            current: The current answer, which is a LLM program fully
                qualified name to validate.

        Returns:
            True if the LLM program fully qualified name is valid, False
            otherwise.

        Raises:
            inquirer.errors.ValidationError: If the LLM program fully
                qualified name is invalid.
        """
        try:
            if not re.match(r".+:.+", current):
                raise inquirer.errors.ValidationError(
                    "",
                    reason="Invalid format.")

            # A valid LLM program fully qualified name should be able to
            # instantiate a `data_model.LazyCallable` object, which will
            # attempt to access the underlying callable object of the LLM
            # program. If it is unable to do so, it will raise an exception.
            data_model.LazyCallable(current)
        except inquirer.errors.ValidationError as error:
            raise error
        except Exception as error:
            # It is not trival identify the relevant Exception subtypes that
            # should be considered actual validation errors, so we catch all
            # exceptions and raise a generic validation error with the
            # exception message appended. The issue with this method is that
            # the validation error message is limited to a single line. If
            # the exception message is too long, it will be truncated. In
            # testing, exception messages seem to be able to fit on a single
            # line, but this should be revisited if we are encountering
            # exceptions with long messages.
            raise inquirer.errors.ValidationError(  # pylint: disable=raise-missing-from
                "", reason=f"Invalid object: {str(error)}")
        return True

    info = ("Enter the fully qualified name (FQN) of the LLM program to test. "
            "The LLM program can be a Python function or a LangChain chain. "
            "The FQN should be in the format "
            "`path.to.python.module:function_or_object_name`.")
    if prefix is not None:
        info = f"[[yellow]{prefix}[/yellow]] {info}"
    rich.print(f"\n{info}")

    return inquirer.text(
        message=("LLM program FQN"),
        validate=validate_llm_program)


def get_new_fully_qualified_name() -> str:
    """Prompt user to enter a new fully qualified name for an LLM program.

    Prompt the user to enter a fully qualified name. Validate the name (see
    `validate_fully_qualified_name` function below). If the name is valid,
    return it. If the name is invalid, continue prompting the user until a
    valid name is entered.

    Returns:
        Fully qualified name for a new LLM program.
    """
    def validate_fully_qualified_name(_: Dict[str, Any], current: str) -> bool:
        """Validate a fully qualified name.

        A fully qualified name is valid if it:
        1. Is in the format <fully qualified module name>:<fully qualified
            object name>.
        2. The file path for the specified module does not already exist.

        Args:
            _: Dictionary containing the answers from previous questions
                (not including the current question). This will be an empty
                dictionary because this question is not part of a collection
                of questions.
            current: The current answer, which is a fully qualified name to
                validate.
        
        Raises:
            inquirer.errors.ValidationError: If the fully qualified name is
                invalid.

        Returns:
            True if the fully qualified name is valid, False otherwise.
        """
        if not re.match(r"\A[^:]+:[^:]+\Z", current):
            raise inquirer.errors.ValidationError(
                "",
                reason="Invalid format.")
        module, _ = current.split(":")
        path_segments = module.split(".")
        path_segments[-1] = path_segments[-1] + ".py"
        path = pathlib.Path(*path_segments)
        if pathlib.Path(path).exists():
            raise inquirer.errors.ValidationError(
                "",
                reason="File already exists.")
        return True

    rich.print("\nEnter the fully qualified name (FQN) for the LLM program "
               "to be created. The FQN should be in the format "
               "\"path.to.python.module:function_name\".")
    return inquirer.text(
        message="LLM program FQN",
        validate=validate_fully_qualified_name)


def choose_new_test_suite() -> bool:
    """Prompt user to create a new test suite or use an existing one.

    Returns:
        True if the user wants to create a new test suite, False if the user
        wants to use an existing test suite.
    """
    create_new, use_existing = ("Create new", "Use existing")
    answer = inquirer.list_input(
        "Create new or use existing test suite?",
        choices=[create_new, use_existing])
    if answer == create_new:
        return True
    return False



# Copyright 2023 Inductor, Inc.
"""Entrypoint for the Inductor CLI."""

import textwrap
from typing import Any, Dict, List, Optional
import pathlib
import re

import inquirer
import rich
from rich import console, syntax
import typer
# Note: we import Annotated from typing_extensions, rather than from typing,
# to enable compatibility with Python 3.8.
from typing_extensions import Annotated
import yaml

from inductor import auth_session, config
from inductor.backend_client import backend_client, wire_model
from inductor.cli import questions
from inductor.data_model import data_model


app = typer.Typer()


def _create_test_suite_file(
    auth_access_token: str,
    default_vars: Optional[Dict[str, Any]] = None,):
    """Prompt user for test suite info to create a test suite file.

    1. Prompt user to create a new test suite or use an existing one.
    2. If creating a new test suite, prompt user for test suite info.
    3. If using an existing test suite, prompt user to select one.
    4. Create a test suite file with the provided info.
    
    Args:
        auth_access_token: Auth0 access token.
        default_vars: Optional dictionary containing default values for
            populating the test suite file.
    """
    if default_vars is None:
        default_vars = {}
    test_suite_args = default_vars

    create_new_test_suite: bool = questions.choose_new_test_suite()
    if create_new_test_suite:
        rich.print("Creating a new test suite file ...")
        has_llm_program_fqn = test_suite_args.get(
            "llm_program_fully_qualified_name") is not None
        question_total = 3 if has_llm_program_fqn else 4
        test_suite_file_path = questions.get_new_file_path(
            "test_suite.yaml",
            [".yml", ".yaml"],
            "test suite",
            prefix=f"1/{question_total}")
        test_suite_args["name"] = questions.get_test_suite_name(
            prefix=f"2/{question_total}")
        test_suite_args["description"] = questions.get_test_suite_description(
            prefix=f"3/{question_total}")

        if not has_llm_program_fqn:
            test_suite_args["llm_program_fully_qualified_name"] = (
                questions.get_llm_program_fully_qualified_name(
                    prefix=f"4/{question_total}"))

        test_suite_args["id"] = None

    else:
        # TODO: This `else` block is a placeholder until the backend is
        # implemented.
        # existing_test_suites = backend_client.get_existing_test_suites()
        # inquirer.list_input(
        #     "Select test suite to use", choices=existing_test_suites)
        # existing_test_suite_args = {}
        # test_suite_args["llm_program_fully_qualified_name"] = "module:object"
        # test_suite_args["id"] = "123"
        # test_suite_args = data_model.merge_test_suite_args(
        #     priority_1_dict=test_suite_args,
        #     priority_2_dict=existing_test_suite_args)
        # raise NotImplementedError("Use existing test suite not implemented.")
        rich.print("To use an existing test suite, run `inductor test "
                   "\[TEST_SUITE_FILE_PATH]`.")  # pylint: disable=anomalous-backslash-in-string
        raise typer.Exit()

    # Get parameter keys from LLM program to populate
    # the test suite file's test case section.
    llm_program = data_model.LazyCallable(
        test_suite_args["llm_program_fully_qualified_name"])
    llm_program_param_keys = llm_program.inputs_signature.keys()
    inputs = {}
    for key in llm_program_param_keys:
        inputs[key] = "<REPLACE WITH VALUE>"

    # Create test suite file contents.
    newline = "\n"
    yaml_string = textwrap.dedent(f"""
    # Test suite file

    # Test suite configuration
    - config:
        id: <WILL BE POPULATED AUTOMATICALLY>
        name: {test_suite_args["name"]}
        # description: {
            test_suite_args["description"]
            if test_suite_args["description"] is not None
            else ""}
        llm_program_fully_qualified_name: {
            test_suite_args["llm_program_fully_qualified_name"]}
        replicas: {
            test_suite_args.get("replicas", 1)
            if test_suite_args.get("replicas", 1) is not None
            else 1}
        
    # A test case (add more `- test:` blocks to add more test cases)
    - test:
        inputs:
          {yaml.dump(inputs, sort_keys=False).strip().replace(
            newline, newline + "          ")}
    
    # Sample quality measure (add more `- quality:` blocks to add more
    # quality measures)
    # - quality:
    #     name: <REPLACE WITH HUMAN-READABLE NAME FOR QUALITY MEASURE>
    #     evaluator: <REPLACE WITH "FUNCTION", "HUMAN", OR "LLM">
    #     evaluation_type: <REPLACE WITH "BINARY" OR "RATING_INT">
    #     # If evaluator is "FUNCTION", spec should be the fully qualified
    #     # name of your quality measure function (e.g., "my.module:my_function").
    #     # If evaluator is "HUMAN", spec should be the instruction or question
    #     # to be posed to a human evaluator. If evaluator is "LLM", spec should
    #     # be in the form of model and prompt (e.g. model: "gpt-3.5-turbo" and
    #     # prompt: "...") or the fully qualified name of an LLM program to be
    #     # used as the evaluator (e.g., "my.module:my_function").  Please see
    #     # the Inductor docs for more details.
    #     spec: <REPLACE WITH VALUE>

    # Sample hyperparameter specification (add more `- hparam:` blocks to add more
    # hyperparameters)
    # - hparam:
    #     name: <REPLACE WITH STRING-VALUED NAME FOR HYPERPARAMETER>
    #     type: <REPLACE WITH "SHORT_STRING", "TEXT", OR "NUMBER">
    #     values: <REPLACE WITH LIST OF VALUES FOR HYPERPARAMETER>
    """).lstrip()

    rich.print("\nTest suite file path:")
    rich.print(test_suite_file_path)
    rich.print("\nTest suite file contents:")
    yaml_string_with_syntax = syntax.Syntax(
        yaml_string, "yaml", theme="ansi_dark", line_numbers=True)
    rich_console = console.Console()
    rich_console.print(yaml_string_with_syntax)

    # Write the test suite file.
    confirm = inquirer.confirm(
        "Create test suite and write contents?",
        default=True)
    if confirm:
        # Create test suite on server if it does not already exist.
        if test_suite_args.get("id") is None:
            response = backend_client.create_test_suite(
                wire_model.CreateTestSuiteRequest(
                    name=test_suite_args["name"],
                    description=test_suite_args.get("description")),
                auth_access_token
            )
            test_suite_args["id"] = response.id
        rich.print("Test suite created.")
        # Replace the `id` field in the test suite file with the ID of the
        # test suite created on the server.
        yaml_string = yaml_string.replace(
            "<WILL BE POPULATED AUTOMATICALLY>", str(test_suite_args["id"]))

        # Write the test suite file.
        with open(test_suite_file_path, "w") as file:
            file.write(yaml_string)
        rich.print("Next, add test cases and quality measures to the file, "
                   "and run your test suite via `inductor test "
                   f"{test_suite_file_path}`")


@app.command("test")
def test(
    test_suite_file_paths: Annotated[
        Optional[List[pathlib.Path]],
        typer.Argument(help=(
            "One or more paths to test suites to run, separated by spaces.")
        )] = None,
    llm_program: Annotated[Optional[str], typer.Option(  # pylint: disable=unused-argument
        "-l",
        "--llm_program",
        help=(
            "Fully qualified name of the Python object (in particular, a "
            "Python function or LangChain chain) representing this LLM "
            "program, in the format: "
            "`<fully qualified module name>:<fully qualified object name>` "
            "(e.g., `my.module:object`). <fully qualified module name> can be "
            "in the format: `path.to.module` or `path/to/module.py`."
    ))] = None,
    replicas: Annotated[Optional[int], typer.Option(  # pylint: disable=unused-argument
        "-r",
        "--replicas",
        help="Number of runs to execute for each test case in test suite(s)."
    )] = None,
    parallelize: Annotated[Optional[int], typer.Option(  # pylint: disable=unused-argument
        "-p",
        "--parallelize",
        help="Maximum number of test cases to run simultaneously in parallel."
    )] = None,
    # TODO: Add support for imports.
    # imports: Annotated[Optional[List[Path]], typer.Option(  # pylint: disable=unused-argument
    #     "-i",
    #     "--imports",
    #     help=("One or more paths to files containing additional test cases "
    #           "or quality measures to include in the test suite(s).")
    # )] = None,
    verbose: Annotated[Optional[bool], typer.Option(
        "-v",
        "--verbose",
        help="Print verbose output.",
    )] = False,
):
    """Run one or more Inductor test suites."""
    local_vars = locals()

    config.verbose = verbose

    # Every time the access token is requested from the session, it is
    # refreshed if it is expired.
    auth_access_token = auth_session.get_auth_session().access_token

    if not test_suite_file_paths:
        _create_test_suite_file(auth_access_token, local_vars)
        raise typer.Exit()

    try:
        test_suites = data_model.get_test_suites_via_cli_args(local_vars)
    except data_model.TestSuiteValueError as error:
        if error.path is not None:
            rich.print(
                f"[bold][red][ERROR][/bold] Error processing test suite file, "
                f"[/red][cyan italic]{error.path}[/cyan italic][red].[/red]"
                f"\n{error.message}"
            )
        else:
            rich.print(
                f"[bold][red][ERROR][/bold] Error processing test suite.[/red]"
                f"\n{error.message}"
            )
        raise typer.Exit()

    multiple_runs = len(test_suites) > 1

    for i, test_suite in enumerate(test_suites):

        if multiple_runs:
            rich.print(rich.rule.Rule(
                f"Test Suite {i+1} of {len(test_suite_file_paths)}",
                style="bold #ffffff"))

        test_suite._run(auth_access_token, prompt_open_results=True)  # pylint: disable=protected-access


def _create_meta_test_suite(
    test_suite_run_id: int,
    quality_measure: wire_model.QualityMeasureWithMetadata,
    llm_quality_measure: Optional[wire_model.QualityMeasure],
    meta_test_suite_name: str,
    meta_test_suite_file_path: pathlib.Path,
    llm_program_fqn: str,
    auth_access_token: str):
    """Create a new meta test suite, and its YAML file.

    Args:
        test_suite_run_id: ID of test suite run based upon which to create a new
            "meta" test suite for meta-evaluation of an LLM-powered quality
            measure.
        quality_measure: The quality measure whose values will serve as target
            values in the meta test suite's test cases.
        llm_quality_measure: The LLM-powered quality measure to be
            meta-evaluated.
        meta_test_suite_name: The name of the new meta test suite.
        meta_test_suite_file_path: The path to the new meta test suite file.
        llm_program_fqn: The fully qualified name of the LLM program
            representing the LLM-powered quality measure to be
            meta-evaluated by the new meta test suite.
        auth_access_token: Auth0 access token.
    """
    # Create test suite on server.
    response = backend_client.create_test_suite(
        wire_model.CreateTestSuiteRequest(name=meta_test_suite_name),
        auth_access_token)

    yaml_string = textwrap.dedent(f"""
    # Test suite file

    # Test suite configuration
    - config:
        id: {response.id}
        name: {meta_test_suite_name}
        llm_program_fully_qualified_name: {llm_program_fqn}
        # Dynamically import new "meta" test cases generated from the following test
        # suite run and quality measure. The new test cases' inputs are based on the
        # given quality measure's inputs (for the given test suite run), and the new
        # test cases' target values are based on the given quality measure's result
        # values (for the given test suite run).
        import:
            - meta_test_suite: "{test_suite_run_id}:{quality_measure.id}"
    """).lstrip()

    if llm_quality_measure is not None:
        yaml_string += "\n"
        yaml_string += textwrap.dedent(f"""
            # Quality measure evaluating if the output of the LLM program (representing
            # the LLM-powered quality measure, {llm_quality_measure.name}) is valid
            # for {llm_quality_measure.evaluation_type} evaluation.
            - quality:
                name: Valid output
                evaluator: FUNCTION
                evaluation_type: BINARY
                spec: inductor.quality:llm_qm_valid_output_{llm_quality_measure.evaluation_type.lower()}
        """).lstrip()

        yaml_string += "\n"
        yaml_string += textwrap.dedent(f"""
            # Quality measure evaluating if the output of the LLM program (representing
            # the LLM-powered quality measure, {llm_quality_measure.name}) is equal to
            # the result of the specified (human-powered) quality measure,
            # {quality_measure.name}, in the test suite run having ID {test_suite_run_id}.
            - quality:
                name: Exact agreement with {quality_measure.name}
                evaluator: FUNCTION
                evaluation_type: BINARY
                spec: inductor.quality:llm_qm_exact_agreement
        """).lstrip()

        if quality_measure.evaluation_type == "RATING_INT":
            yaml_string += "\n"
            yaml_string += textwrap.dedent(f"""
            # Quality measure evaluating if the output of the LLM program
            # (representing the LLM-powered quality measure,
            # {llm_quality_measure.name}) is within +/- 1 rating point of the
            # result value of the specified (human-powered) quality measure,
            # {quality_measure.name}, in the test suite run having ID {test_suite_run_id}.
            - quality:
                name: Plus-or-minus-one agreement with {quality_measure.name}
                evaluator: FUNCTION
                evaluation_type: BINARY
                spec: inductor.quality:llm_qm_plus_minus_one_agreement
            """).lstrip()

    # If ancestor directories do not exist, create them.
    meta_test_suite_file_path.parent.mkdir(parents=True, exist_ok=True)

    # Write the test suite file.
    with open(meta_test_suite_file_path, "w") as file:
        file.write(yaml_string)


def _create_meta_llm_program_file(
    llm_program_fqn: str,
    llm_quality_measure_spec: Optional[Dict[str, str]] = None
):
    """Create a new LLM program file for testing an LLM quality measure.
    
    Args:
        llm_program_fqn: The fully qualified name of the LLM program to create.
        llm_quality_measure_spec: The spec of the LLM-powered quality measure.
    """
    llm_program_path, llm_program_function_name = llm_program_fqn.split(":")
    llm_program_path = llm_program_path.replace(".", "/")
    llm_program_path = pathlib.Path(f"{llm_program_path}.py")

    import_openai = (
        "import openai\n" if llm_quality_measure_spec is not None else "")

    python_string = f"""
        \"\"\"This file was automatically generated by Inductor. You can modify
        it as needed.\"\"\"

        from typing import Any, Dict

        import inductor
        {import_openai}

        def {llm_program_function_name}(
            llm_program_output: str,
            llm_program_inputs: Dict[str, Any]
        ) -> str:
            \"\"\"Evaluates the quality of an LLM program's output.

            Args:
                llm_program_output: Output of the LLM program.
                llm_program_inputs: Inputs to the LLM program.

            Returns:
                The value produced by this LLM-powered quality measure.
            \"\"\"
    """

    if llm_quality_measure_spec is not None:
        prompt = llm_quality_measure_spec["prompt"]
        prompt_lines = prompt.split("\n")
        prompt_lines = [
            line.replace("\"", "\\\"") for line in prompt_lines]
        prompt = "\n" + "\n".join(
            [f"                \"{line}\"" for line in prompt_lines])
        python_string += f"""
            client = openai.OpenAI()
            prompt = ({prompt})
            model = "{llm_quality_measure_spec["model"]}"

            execution = {{
                "output": llm_program_output,
                "inputs": llm_program_inputs
            }}
            chat_completion = client.chat.completions.create(
                messages=[
                    {{
                        "role": "system",
                        "content": prompt.format_map(execution)
                    }},
                ],
                model=model)
            return chat_completion.choices[0].message.content
        """
    else:
        python_string += (
            "        # Add code here "
            "to implement this LLM-powered quality measure\n")

    python_string = textwrap.dedent(python_string).lstrip()

    # If ancestor directories do not exist, create them.
    llm_program_path.parent.mkdir(parents=True, exist_ok=True)

    # Write the LLM program file.
    with open(llm_program_path, "w") as file:
        file.write(python_string)


@app.command("create-meta-test-suite")
def create_meta_test_suite(
    meta_test_suite_spec: Annotated[str, typer.Argument(
        help="Specification of meta test suite to be used for "
        "evaluating an LLM-powered quality measure, in the "
        "format \"<test suite run ID>:<quality measure ID>\".  "
        "The specified (human-powered) quality measure's "
        "outputs (in the specified test suite run) will be used "
        "as target values within the new meta test suite."
    )],
    llm_quality_measure_id: Annotated[Optional[int], typer.Option(
        "-l",
        "--llm_quality_measure",
        help="ID of an LLM-powered quality measure (in the test suite "
             "run given by `meta_test_suite_spec`), from which a new LLM "
             "program will be created."
    )] = None,
    verbose: Annotated[Optional[bool], typer.Option(
        "-v",
        "--verbose",
        help="Print verbose output.",
    )] = False
):
    """Create a new test suite based on an existing test suite run in order to
    "meta-evaluate" an LLM-powered quality measure and ensure that it aligns
    with an existing human-powered quality measure.  Please see Inductor's docs
    for more information."""
    config.verbose = verbose

    auth_access_token = auth_session.get_auth_session().access_token

    # Validate and parse meta_test_suite_spec argument.
    try:
        if len(list(map(int, meta_test_suite_spec.split(":")))) != 2:
            raise ValueError
    except (TypeError, ValueError) as error:
        rich.print(
            "meta_test_suite_spec argument must be in the format "
            "<test suite run ID>:<quality measure ID>.")
        raise typer.Exit() from error
    test_suite_run_id, quality_measure_id = list(
        map(int, meta_test_suite_spec.split(":")))

    # Get specified quality measures
    quality_measure_ids_to_get = [quality_measure_id]
    if llm_quality_measure_id is not None:
        quality_measure_ids_to_get.append(llm_quality_measure_id)
    test_suite_run_quality_measures_response = (
        backend_client.get_test_suite_run_quality_measures(
            test_suite_run_id,
            wire_model.GetTestSuiteRunQualityMeasuresRequest(
                quality_measure_ids=quality_measure_ids_to_get),
            auth_access_token))
    quality_measure = (
        test_suite_run_quality_measures_response.quality_measures[0])
    llm_quality_measure = (
        test_suite_run_quality_measures_response.quality_measures[1]
        if llm_quality_measure_id is not None
        else None)

    if quality_measure is None:
        rich.print("Quality measure not found.")
        raise typer.Exit()

    # Remove all non-alphanumeric and non-underscore characters, replace
    # spaces with underscores, and convert all characters to lowercase.
    quality_measure_name_underscored = re.sub(
        r"[^a-zA-Z0-9_]", "", quality_measure.name.replace(" ", "_")).lower()

    new_meta_test_suite_name = questions.get_test_suite_name(
        quality_measure_name_underscored + "_meta_test_suite",
        "meta test suite")

    meta_test_suite_file_path = questions.get_new_file_path(
        "meta_test_suite.yaml",
        [".yml", ".yaml"],
        "meta test suite",)

    llm_program_fqn = None
    llm_quality_measure_spec = None

    if llm_quality_measure_id is not None:
        if llm_quality_measure is None:
            rich.print("LLM-powered quality measure not found.")
            raise typer.Exit()

        if llm_quality_measure.evaluator != "LLM":
            rich.print(
                "Quality measure specified by "
                "llm_quality_measure_id is not LLM-powered.")
            raise typer.Exit()

        if isinstance(llm_quality_measure.spec, str):
            llm_program_fqn = llm_quality_measure.spec
        else:
            llm_quality_measure_spec = llm_quality_measure.spec

    if llm_program_fqn is None:
        llm_program_fqn = questions.get_new_fully_qualified_name()

    # Write the test suite file and LLM program file.
    create_meta_llm_program_file = (
        llm_quality_measure is None or
        not isinstance(llm_quality_measure.spec, str))
    info = "\n[[yellow]?[/yellow]] Create meta test suite file"
    if create_meta_llm_program_file:
        info += " and LLM program file,"
    rich.print(info)
    confirm = inquirer.confirm("and write contents?", default=True)
    if confirm:
        if create_meta_llm_program_file:
            _create_meta_llm_program_file(
                llm_program_fqn,
                llm_quality_measure_spec)

        _create_meta_test_suite(
            test_suite_run_id,
            quality_measure,
            llm_quality_measure or wire_model.QualityMeasure(
                name=llm_program_fqn,
                evaluator="LLM",
                evaluation_type=quality_measure.evaluation_type,
                spec=llm_program_fqn),
            new_meta_test_suite_name,
            meta_test_suite_file_path,
            llm_program_fqn,
            auth_access_token)

        meta_test_suite_file_path = meta_test_suite_file_path.resolve()
        if create_meta_llm_program_file:
            rich.print("New LLM program for quality measure meta-evaluation "
                    f"created in [green]{llm_program_fqn}[/green].")
        rich.print("New test suite for quality measure meta-evaluation created "
                   f"in {meta_test_suite_file_path}.")
        rich.print("To run your meta test suite, execute the command `inductor "
                   f"test {meta_test_suite_file_path}`")



"""This module provides pre-built quality measures (QMs)."""

from typing import Any, Union

from inductor.data_model import data_model


def llm_qm_valid_output_binary(output: Any) -> bool:
    """Checks if an LLM QM's output is valid for BINARY evaluation_type.

    Args:
        output: The output of the execution of an LLM-powered quality measure.

    Returns:
        True if output is in {"Y", "N", "YES", "NO", True, False}, and False
        otherwise.
    """
    return output in {"Y", "N", "YES", "NO", True, False}


def llm_qm_valid_output_rating_int(output: Any) -> bool:
    """Checks if an LLM QM's output is valid for RATING_INT evaluation_type.

    Args:
        output: The output of the execution of an LLM-powered quality measure.

    Returns:
        True if output is in {"1", "2", "3", "4", "5", 1, 2, 3, 4, 5}, and False
        otherwise.
    """
    return output in {"1", "2", "3", "4", "5", 1, 2, 3, 4, 5}


def llm_qm_exact_agreement(
    output: Union[str, int, bool],
    _,
    test_case: data_model.TestCase
) -> bool:
    """Checks if an LLM quality measure's output is equal to its target value.

    This quality measure is intended for use in "meta" evaluation of an
    LLM quality measure.

    Args:
        output: The output of the execution of an LLM-powered quality measure.
        test_case: The "meta" test case on which the LLM-powered quality
            measure was executed.
    
    Returns:
        True if the output matches the target value given by the test_case,
        False otherwise.
    """
    def _normalize_output(
        output: Union[str, int, bool]
    ) -> Union[str, int, bool]:
        if isinstance(output, str):
            output_normalized = output.strip().upper()
            if output_normalized in ("YES", "Y"):
                return True
            elif output_normalized in ("NO", "N"):
                return False
            elif output_normalized.isdigit():
                return int(output_normalized)
        return output
    return _normalize_output(output) == _normalize_output(test_case.output)


def llm_qm_plus_minus_one_agreement(
    output: Union[str, int, bool],
    _,
    test_case: data_model.TestCase
) -> bool:
    """Checks if an LLM quality measure's output is within +/- 1 of target.

    This quality measure is intended for use in "meta" evaluation of an
    LLM quality measure whose evaluation type is RATING_INT.

    Args:
        output: The output of the execution of an LLM-powered quality measure.
        test_case: The "meta" test case on which the LLM-powered quality
            measure was executed.
    
    Returns:
        True if the output is within +/- 1 of the target value given by
        the test_case, False otherwise.
    """
    output = int(output)
    test_case_output = int(test_case.output)
    return test_case_output - 1 <= output <= test_case_output + 1



# Copyright 2023 Inductor, Inc.
"""Functions for interfacing with the Inductor backend."""

from typing import Any, Dict, Literal, Optional, Union

import httpx
import pydantic

from inductor import config
from inductor.backend_client import wire_model


# TODO: Cleanup these globals.
_MOCK_ENDPOINTS = False


def _send_request(
    method: Literal["get", "post", "put", "delete"],
    endpoint: str,
    auth_access_token: str,
    request_body: Optional[Union[pydantic.BaseModel, Dict[str, Any]]] = None
) -> httpx.Response:
    """Sends a request to the backend with the given method and endpoint.

    Args:
        method: The HTTP method to use. Can be one of "get", "post", "put", or
            "delete".
        endpoint: The endpoint to send the request to. This should not begin
            with a slash.
        auth_access_token: Auth0 access token. This is included in the request
            header.
        request_body: The request body to send. If get or delete, the request
            body is ignored.

    Returns:
        The response from the backend.

    Raises:
        RuntimeError: If the response status code is not a successful response
            (i.e., if the response status code is not in the range (200 – 299)).
    """
    if isinstance(request_body, pydantic.BaseModel):
        request_body_json = request_body.model_dump(mode="json")
    else:
        request_body_json = request_body

    response = httpx.request(
        method,
        f"{config.settings.inductor_api_url}/api/client/v1/{endpoint}",
        headers={
            "Authorization": f"Bearer {auth_access_token}",
            **(config.settings.custom_request_headers or {}),
        },
        json=request_body_json,
    )
    # TODO: #433 - These print statements should be available in --verbose
    # mode.
    # print(f"Request: {request_body_json}")
    # print(f"Response: {response.text}")
    # print(f"Status code: {response.status_code}")
    # print(f"JSON: {response.json()}")
    if response.status_code < 200 or response.status_code >= 300:
        raise RuntimeError(
            f"{method} request to `{endpoint}` failed with status code "
            f"{response.status_code} and response text: {response.text}.")
    return response


def create_api_key(
    request: wire_model.CreateApiKeyRequest,
    auth_access_token: str):
    """POST request to create an API key.
    
    Args:
        request: CreateApiKeyRequest object
        auth_access_token: Auth0 access token.
    """
    if _MOCK_ENDPOINTS:
        return
    _send_request("post", "create-api-key", auth_access_token, request)


def create_test_suite(
    request: wire_model.CreateTestSuiteRequest,
    auth_access_token: str
) -> wire_model.CreateTestSuiteResponse:
    """POST request to create a test suite.

    Args:
        request: CreateTestSuiteRequest object
        auth_access_token: Auth0 access token.
    
    Returns:
        CreateTestSuiteResponse object.
    """
    if _MOCK_ENDPOINTS:
        return wire_model.CreateTestSuiteResponse(id=123)
    response = _send_request(
        "post", "create-test-suite", auth_access_token, request)
    return wire_model.CreateTestSuiteResponse(**response.json())


def create_test_suite_run(
    request: wire_model.CreateTestSuiteRunRequest,
    auth_access_token: str
) -> wire_model.CreateTestSuiteRunResponse:
    """POST request to create a test suite run.
    
    Args:
        request: CreateTestSuiteRunRequest object
        auth_access_token: Auth0 access token.
    
    Returns:
        CreateTestSuiteRunResponse object.
    """
    if _MOCK_ENDPOINTS:
        return wire_model.CreateTestSuiteRunResponse(
            test_suite_run_id=123,
            test_case_ids=[1, 2, 3, 4, 5, 6],
            quality_measure_ids=[7, 8, 9, 10, 11, 12],
            hparam_spec_ids=[13, 14, 15, 16, 17, 18],
            url="http://localhost:5000/MOCK_URL",
        )
    response = _send_request(
        "post", "create-test-suite-run", auth_access_token, request)
    return wire_model.CreateTestSuiteRunResponse(**response.json())


def log_test_case_execution(
    request: wire_model.LogTestCaseExecutionRequest,
    auth_access_token: str):
    """POST request to log test case execution.

    Args:
        request: LogTestCaseExecutionRequest object
        auth_access_token: Auth0 access token.
    """
    if _MOCK_ENDPOINTS:
        return
    _send_request(
        "post", "log-test-case-execution", auth_access_token, request)


def complete_test_suite_run(
    request: wire_model.CompleteTestSuiteRunRequest,
    auth_access_token: str):
    """POST request to complete test suite run.

    Args:
        request: CompleteTestSuiteRunRequest object
        auth_access_token: Auth0 access token.
    """
    if _MOCK_ENDPOINTS:
        return
    _send_request(
        "post", "complete-test-suite-run", auth_access_token, request)


def log_llm_program_execution(
    request: wire_model.LogLlmProgramExecutionRequest,
    auth_access_token: str
) -> wire_model.LogLlmProgramExecutionResponse:
    """POST request to log LLM program execution.

    Args:
        request: LogLlmProgramExecutionRequest object
        auth_access_token: Auth0 access token.

    Returns:
        LogLlmProgramExecutionResponse object.
    """
    if _MOCK_ENDPOINTS:
        return wire_model.LogLlmProgramExecutionResponse(id=123)
    response = _send_request(
        "post", "log-llm-program-execution", auth_access_token, request)
    return wire_model.LogLlmProgramExecutionResponse(**response.json())


def create_live_deployment(
    request: wire_model.CreateLiveDeploymentRequest,
    auth_access_token: str) -> wire_model.CreateLiveDeploymentResponse:
    """POST request to create live deployment.

    Args:
        request: `CreateLiveDeploymentRequest` object.
        auth_access_token: Auth0 access token.

    Returns:
        `CreateLiveDeploymentResponse` object.
    """
    response = _send_request(
        "post", "create-live-deployment", auth_access_token, request)
    return wire_model.CreateLiveDeploymentResponse(**response.json())


def get_test_suite_run_quality_measures(
    test_suite_run_id: int,
    request: wire_model.GetTestSuiteRunQualityMeasuresRequest,
    auth_access_token: str
) -> wire_model.GetTestSuiteRunQualityMeasuresResponse:
    """POST request to get a test suite run's quality measures.

    Args:
        test_suite_run_id: Test suite run ID.
        request: `GetTestSuiteRunQualityMeasuresRequest` object.
        auth_access_token: Auth0 access token.

    Returns:
        `GetTestSuiteRunQualityMeasuresResponse` object.
    """
    response = _send_request(
        "post",
        f"test-suite-run/{test_suite_run_id}/get-quality-measures",
        auth_access_token,
        request)
    return wire_model.GetTestSuiteRunQualityMeasuresResponse(**response.json())


def log_values_for_execution(
    request: wire_model.LogValuesForExecutionRequest,
    auth_access_token: str):
    """POST request to log values for an execution.

    Args:
        request: LogValuesForExecutionRequest object
        auth_access_token: Auth0 access token.
    """
    if _MOCK_ENDPOINTS:
        return
    _send_request(
        "post", "log-values-for-execution", auth_access_token, request)
